@@ -31,16 +31,17 @@ class DNSError(Exception):
 # DNS codes
 
 QTYPE =  Bimap('QTYPE',
-                {1:'A', 2:'NS', 5:'CNAME', 6:'SOA', 12:'PTR', 13:'HINFO',
+        {1:'A', 2:'NS', 5:'CNAME', 6:'SOA', 10:'NULL', 12:'PTR', 13:'HINFO',
                     15:'MX', 16:'TXT', 17:'RP', 18:'AFSDB', 24:'SIG', 25:'KEY',
                     28:'AAAA', 29:'LOC', 33:'SRV', 35:'NAPTR', 36:'KX',
                     37:'CERT', 38:'A6', 39:'DNAME', 41:'OPT', 42:'APL',
                     43:'DS', 44:'SSHFP', 45:'IPSECKEY', 46:'RRSIG', 47:'NSEC',
                     48:'DNSKEY', 49:'DHCID', 50:'NSEC3', 51:'NSEC3PARAM',
-                    52:'TLSA', 55:'HIP', 59:'CDS', 60:'CDNSKEY',
-                    61:'OPENPGPKEY',99:'SPF', 249:'TKEY', 250:'TSIG',
-                    251:'IXFR', 252:'AXFR', 255:'ANY', 256:'URI', 257:'CAA',
-                    32768:'TA', 32769:'DLV'}, DNSError)
+                    52:'TLSA', 53:'HIP', 55:'HIP', 59:'CDS', 60:'CDNSKEY',
+                    61:'OPENPGPKEY', 62:'CSYNC', 63:'ZONEMD', 64:'SVCB',
+                    65:'HTTPS', 99:'SPF', 108:'EUI48', 109:'EUI64', 249:'TKEY',
+                    250:'TSIG', 251:'IXFR', 252:'AXFR', 255:'ANY', 256:'URI',
+                    257:'CAA', 32768:'TA', 32769:'DLV'}, DNSError)
 
 CLASS =  Bimap('CLASS',
                 {1:'IN', 2:'CS', 3:'CH', 4:'Hesiod', 254:'None', 255:'*'},
@@ -355,28 +356,32 @@ class DNSRecord(object):
             inet = socket.AF_INET6
         else:
             inet = socket.AF_INET
-        if tcp:
-            if len(data) > 65535:
-                raise ValueError("Packet length too long: %d" % len(data))
-            data = struct.pack("!H",len(data)) + data
-            sock = socket.socket(inet,socket.SOCK_STREAM)
-            if timeout is not None:
-                sock.settimeout(timeout)
-            sock.connect((dest,port))
-            sock.sendall(data)
-            response = sock.recv(8192)
-            length = struct.unpack("!H",bytes(response[:2]))[0]
-            while len(response) - 2 < length:
-                response += sock.recv(8192)
-            sock.close()
-            response = response[2:]
-        else:
-            sock = socket.socket(inet,socket.SOCK_DGRAM)
-            if timeout is not None:
-                sock.settimeout(timeout)
-            sock.sendto(self.pack(),(dest,port))
-            response,server = sock.recvfrom(8192)
-            sock.close()
+        try:
+            sock = None
+            if tcp:
+                if len(data) > 65535:
+                     raise ValueError("Packet length too long: %d" % len(data))
+                data = struct.pack("!H",len(data)) + data
+                sock = socket.socket(inet,socket.SOCK_STREAM)
+                if timeout is not None:
+                    sock.settimeout(timeout)
+                sock.connect((dest,port))
+                sock.sendall(data)
+                response = sock.recv(8192)
+                length = struct.unpack("!H",bytes(response[:2]))[0]
+                while len(response) - 2 < length:
+                    response += sock.recv(8192)
+                response = response[2:]
+            else:
+                sock = socket.socket(inet,socket.SOCK_DGRAM)
+                if timeout is not None:
+                    sock.settimeout(timeout)
+                sock.sendto(self.pack(),(dest,port))
+                response,server = sock.recvfrom(8192)
+        finally:
+            if (sock is not None):
+                sock.close()
+
         return response
 
     def format(self,prefix="",sort=False):
@@ -1324,6 +1329,9 @@ class PTR(CNAME):
 class NS(CNAME):
     pass
 
+class DNAME(CNAME):
+    pass
+
 class SOA(RD):
 
     times = ntuple_range('times',5,0,4294967295)
@@ -1621,23 +1629,35 @@ def decode_type_bitmap(type_bitmap):
 def encode_type_bitmap(rrlist):
     """
         Encode RR type bitmap in NSEC record
-        XXX Currenly support window #0 (RR 1-255)
 
         >>> p = lambda x: print(binascii.hexlify(x).decode())
         >>> p(encode_type_bitmap(['A','TXT','AAAA','RRSIG','NSEC']))
         0006400080080003
         >>> p(encode_type_bitmap(['A','NS','SOA','TXT','AAAA','RRSIG','NSEC','DNSKEY']))
         000762008008000380
+        >>> p(encode_type_bitmap(['A','ANY','URI','CAA','TA','DLV']))
+        002040000000000000000000000000000000000000000000000000000000000000010101c08001c0
     """
-    bitmap = bytearray([0]*32)
-    for rr in rrlist:
-        v = getattr(QTYPE,rr)
-        if v > 255:
-            raise ValueError("Only support window #0 (RR 1-255) [%s]" % rr)
+    rrlist = sorted([getattr(QTYPE,rr) for rr in rrlist])
+    buf = DNSBuffer()
+    curWindow = rrlist[0]//256
+    bitmap = bytearray(32)
+    n = len(rrlist)-1
+    for i, rr in enumerate(rrlist):
+        v = rr - curWindow*256
         bitmap[v//8] |= 1 << (7 - v%8)
-    while bitmap[-1] == 0:
-        bitmap = bitmap[:-1]
-    return struct.pack("BB",0,len(bitmap)) + bitmap
+
+        if i == n or rrlist[i+1] >= (curWindow+1)*256:
+            while bitmap[-1] == 0:
+                bitmap = bitmap[:-1]
+            buf.pack("BB", curWindow, len(bitmap))
+            buf.append(bitmap)
+
+            if i != n:
+                curWindow = rrlist[i+1]//256
+                bitmap = bytearray(32)
+
+    return buf.data
 
 class NSEC(RD):
 
